@@ -8,21 +8,24 @@
     :copyright: (c) 2013 by Hsiaoming Yang.
 """
 
-import urllib2
-import logging
 import oauthlib.oauth1
 import oauthlib.oauth2
 from functools import wraps
-from oauthlib.common import to_unicode
-from urlparse import urljoin, urlparse
-from flask import request, redirect, json, session
-from werkzeug import url_quote, url_decode, url_encode, parse_options_header
+from oauthlib.common import to_unicode, PY3
+from flask import request, redirect, json, session, current_app
+from werkzeug import url_quote, url_decode, url_encode
+from werkzeug import parse_options_header, cached_property
+try:
+    from urlparse import urljoin, urlparse
+    import urllib2 as http
+except ImportError:
+    from urllib import request as http
+    from urllib.parse import urljoin, urlparse
+from ._utils import log
 
 
 __all__ = ('OAuth', 'OAuthRemoteApp', 'OAuthResponse', 'OAuthException')
 
-log = logging.getLogger('flask_oauthlib')
-
 
 class OAuth(object):
     """Registry for remote applications.
@@ -149,11 +152,11 @@ def make_request(uri, headers=None, data=None, method=None,
         resp.code = resp.status_code
         return resp, resp.data
 
-    req = urllib2.Request(uri, headers=headers, data=data)
+    req = http.Request(uri, headers=headers, data=data)
     req.get_method = lambda: method.upper()
     try:
-        resp = urllib2.urlopen(req)
-    except urllib2.HTTPError as resp:
+        resp = http.urlopen(req)
+    except http.HTTPError as resp:
         pass
     content = resp.read()
     resp.close()
@@ -198,6 +201,8 @@ class OAuthException(RuntimeError):
         self.data = data
 
     def __str__(self):
+        if PY3:
+            return self.message
         return self.message.encode('utf-8')
 
     def __unicode__(self):
@@ -226,43 +231,121 @@ class OAuthRemoteApp(object):
     :param content_type: force to parse the content with this content_type,
                          usually used when the server didn't return the
                          right content type.
+
+    .. versionadded:: 0.3.0
+
+    :param app_key: lazy load configuration from Flask app config with
+                    this app key
     """
     def __init__(
         self, oauth, name,
-        base_url, request_token_url,
-        access_token_url, authorize_url,
-        consumer_key, consumer_secret,
+        base_url=None,
+        request_token_url=None,
+        access_token_url=None,
+        authorize_url=None,
+        consumer_key=None,
+        consumer_secret=None,
         request_token_params=None,
         access_token_params=None,
         access_token_method='GET',
         content_type=None,
         test_client=None,
+        app_key=None,
         encoding='utf-8',
     ):
-
         self.oauth = oauth
-        self.base_url = base_url
         self.name = name
-        self.request_token_url = request_token_url
-        self.access_token_url = access_token_url
-        self.authorize_url = authorize_url
-        self.consumer_key = consumer_key
-        self.consumer_secret = consumer_secret
-        self.request_token_params = request_token_params or {}
-        self.access_token_params = access_token_params or {}
-        self.access_token_method = access_token_method
-        self.content_type = content_type
-        self.test_client = test_client
-        self.encoding = encoding
+
+        if (not consumer_key or not consumer_secret) and not app_key:
+            raise TypeError(
+                'OAuthRemoteApp requires consumer key and secret'
+            )
+
+
+        self._base_url = base_url
+        self._request_token_url = request_token_url
+        self._access_token_url = access_token_url
+        self._authorize_url = authorize_url
+        self._consumer_key = consumer_key
+        self._consumer_secret = consumer_secret
+        self._request_token_params = request_token_params or {}
+        self._access_token_params = access_token_params or {}
+        self._access_token_method = access_token_method
+        self._content_type = content_type
 
         self._tokengetter = None
 
+        self.test_client = test_client
+        self.app_key = app_key
+        self.encoding = encoding
+
+    @cached_property
+    def base_url(self):
+        return self._get_property('base_url')
+
+    @cached_property
+    def request_token_url(self):
+        return self._get_property('request_token_url', None)
+
+    @cached_property
+    def access_token_url(self):
+        return self._get_property('access_token_url')
+
+    @cached_property
+    def authorize_url(self):
+        return self._get_property('authorize_url')
+
+    @cached_property
+    def consumer_key(self):
+        return self._get_property('consumer_key')
+
+    @cached_property
+    def consumer_secret(self):
+        return self._get_property('consumer_secret')
+
+    @cached_property
+    def request_token_params(self):
+        return self._get_property('request_token_params', None)
+
+    @cached_property
+    def access_token_params(self):
+        return self._get_property('access_token_params', None)
+
+    @cached_property
+    def access_token_method(self):
+        return self._get_property('access_token_method', 'GET')
+
+    @cached_property
+    def content_type(self):
+        return self._get_property('content_type', None)
+
+    def _get_property(self, key, default=False):
+        attr = getattr(self, '_%s' % key)
+        if attr:
+            return attr
+        if default is not False and not self.app_key:
+            return attr
+        app = self.oauth.app or current_app
+        config = app.config[self.app_key]
+        if default is not False:
+            return config.get(key, default)
+        return config[key]
+
     def make_client(self, token=None):
         # request_token_url is for oauth1
         if self.request_token_url:
             client = oauthlib.oauth1.Client(
                 self.consumer_key, self.consumer_secret
             )
+
+            params = self.request_token_params
+            if 'signature_method' in params:
+                client.signature_method = _encode(params['signature_method'])
+            if 'rsa_key' in params:
+                client.rsa_key = _encode(params['rsa_key'])
+            if 'signature_type' in params:
+                client.signature_type = _encode(params['signature_type'])
+
             if token and isinstance(token, (tuple, list)):
                 client.resource_owner_key, client.resource_owner_secret = token
         else:
@@ -405,9 +488,15 @@ class OAuthRemoteApp(object):
 
         client = self.make_client()
         client.callback_uri = _encode(callback, self.encoding)
+
+        realm = self.request_token_params.get('realm')
+        realms = self.request_token_params.get('realms')
+        if not realm and realms:
+            realm = ' '.join(realms)
         uri, headers, _ = client.sign(
-            self.expand_url(self.request_token_url)
+            self.expand_url(self.request_token_url), realm=realm
         )
+        log.debug('Generate request token header %r', headers)
         resp, content = make_request(
             uri, headers, test_client=self.test_client
         )
