@@ -9,21 +9,18 @@
 """
 
 import os
-import logging
 import datetime
 from functools import wraps
-from flask import _app_ctx_stack
 from flask import request, url_for
 from flask import redirect, make_response, abort
 from werkzeug import cached_property
 from oauthlib import oauth2
 from oauthlib.oauth2 import RequestValidator, Server
 from oauthlib.common import to_unicode
+from .._utils import _extract_params, log, decode_base64
 
 __all__ = ('OAuth2Provider', 'OAuth2RequestValidator')
 
-log = logging.getLogger('flask_oauthlib')
-
 
 class OAuth2Provider(object):
     """Provide secure services using OAuth2.
@@ -64,11 +61,13 @@ class OAuth2Provider(object):
 
         @app.route('/api/user')
         @oauth.require_oauth('email', 'username')
-        def user():
-            return jsonify(g.user)
+        def user(oauth):
+            return jsonify(oauth.user)
     """
 
     def __init__(self, app=None):
+        self._before_request_funcs = []
+        self._after_request_funcs = []
         if app:
             self.init_app(app)
 
@@ -146,6 +145,40 @@ class OAuth2Provider(object):
             return Server(validator, token_expires_in=expires_in)
         raise RuntimeError('application not bound to required getters')
 
+    def before_request(self, f):
+        """Register functions to be invoked before accessing the resource.
+
+        The function accepts nothing as parameters, but you can get
+        information from `Flask.request` object. It is usually useful
+        for setting limitation on the client request::
+
+            @oauth.before_request
+            def limit_client_request():
+                client_id = request.values.get('client_id')
+                if not client_id:
+                    return
+                client = Client.get(client_id)
+                if over_limit(client):
+                    return abort(403)
+
+                track_request(client)
+        """
+        self._before_request_funcs.append(f)
+
+    def after_request(self, f):
+        """Register functions to be invoked after accessing the resource.
+
+        The function accepts ``valid`` and ``request`` as parameters,
+        and it should return a tuple of them::
+
+            @oauth.after_request
+            def valid_after_request(valid, oauth):
+                if oauth.user in black_list:
+                    return False, oauth
+                return valid, oauth
+        """
+        self._after_request_funcs.append(f)
+
     def clientgetter(self, f):
         """Register a function as the client getter.
 
@@ -370,11 +403,18 @@ class OAuth2Provider(object):
         def wrapper(f):
             @wraps(f)
             def decorated(*args, **kwargs):
+                for func in self._before_request_funcs:
+                    func()
+
                 server = self.server
                 uri, http_method, body, headers = _extract_params()
                 valid, req = server.verify_request(
                     uri, http_method, body, headers, scopes
                 )
+
+                for func in self._after_request_funcs:
+                    valid, req = func(valid, req)
+
                 if not valid:
                     return abort(403)
                 return f(*((req,) + args), **kwargs)
@@ -411,8 +451,8 @@ class OAuth2RequestValidator(RequestValidator):
         log.debug('Authenticate client %r', auth)
         if auth:
             try:
-                _, base64 = auth.split(' ')
-                client_id, client_secret = base64.decode('base64').split(':')
+                _, s = auth.split(' ')
+                client_id, client_secret = decode_base64(s).split(':')
                 client_id = to_unicode(client_id, 'utf-8')
                 client_secret = to_unicode(client_secret, 'utf-8')
             except Exception as e:
@@ -432,11 +472,7 @@ class OAuth2RequestValidator(RequestValidator):
             log.debug('Authenticate client failed, secret not match.')
             return False
 
-        confidential = 'confidential'
-        if hasattr(client, 'confidential'):
-            confidential = client.confidential
-
-        if client.client_type != confidential:
+        if client.client_type != 'confidential':
             log.debug('Authenticate client failed, not confidential.')
             return False
         log.debug('Authenticate client success.')
@@ -459,11 +495,7 @@ class OAuth2RequestValidator(RequestValidator):
 
         # authenticate non-confidential client_type only
         # most of the clients are of public client_type
-        confidential = 'confidential'
-        if hasattr(client, 'confidential'):
-            confidential = client.confidential
-
-        if client.client_type == confidential:
+        if client.client_type == 'confidential':
             log.debug('Authenticate client failed, confidential client.')
             return False
         return True
@@ -494,10 +526,25 @@ class OAuth2RequestValidator(RequestValidator):
 
         return grant.redirect_uri == redirect_uri
 
+    def get_original_scopes(self, refresh_token, request, *args, **kwargs):
+        """Get the list of scopes associated with the refresh token.
+
+        This method is used in the refresh token grant flow.  We return
+        the scope of the token to be refreshed so it can be applied to the
+        new access token.
+        """
+        log.debug('Obtaining scope of refreshed token.')
+        tok = self._tokengetter(refresh_token=refresh_token)
+        return tok.scopes
+
     def confirm_scopes(self, refresh_token, scopes, request, *args, **kwargs):
         """Ensures the requested scope matches the scope originally granted
         by the resource owner. If the scope is omitted it is treated as equal
-        to the scope originally granted by the resource owner
+        to the scope originally granted by the resource owner.
+
+        DEPRECATION NOTE: This method will cease to be used in oauthlib>0.4.2,
+        future versions of ``oauthlib`` use the validator method
+        ``get_original_scopes`` to determine the scope of the refreshed token.
         """
         if not scopes:
             log.debug('Scope omitted for refresh token %r', refresh_token)
@@ -718,20 +765,3 @@ class OAuth2RequestValidator(RequestValidator):
             return False
         log.debug('Password credential authorization is disabled.')
         return False
-
-
-def _extract_params():
-    """Extract request params."""
-    log.debug('Extracting parameters from request.')
-    uri = request.url
-    http_method = request.method
-    headers = dict(request.headers)
-    if 'wsgi.input' in headers:
-        del headers['wsgi.input']
-    if 'wsgi.errors' in headers:
-        del headers['wsgi.errors']
-    if 'Http-Authorization' in headers:
-        headers['Authorization'] = headers['Http-Authorization']
-
-    body = request.form.to_dict()
-    return uri, http_method, body, headers
